/**
  *  \file planetscentral/host/exec.cc
  */

#include <fstream>
#include <stdexcept>
#include <map>
#include <time.h>
#include "cpluslib/format.h"
#include "game/struct.h"
#include "io/io.h"
#include "io/streams.h"
#include "planetscentral/dbclient/dbclient.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/host/db.h"
#include "planetscentral/host/exec.h"
#include "planetscentral/host/game.h"
#include "planetscentral/host/host.h"
#include "planetscentral/host/import.h"
#include "planetscentral/host/runner.h"
#include "planetscentral/host/schedule.h"
#include "cpluslib/misc.h"
#include "planetscentral/dbclient/fileclient.h"
#include "planetscentral/host/install.h"
#include "planetscentral/dbclient/routerclient.h"
#include "planetscentral/host/victory.h"
#include "planetscentral/host/ranking.h"
#include "planetscentral/dbclient/mailclient.h"
#include "planetscentral/host/newgame.h"

static string_t
makeGameUrl(int32_t gameId, string_t gameName)
{
    string_t raw = format("%d-%s", gameId, gameName);
    string_t result;
    bool needDash = false;
    for (string_t::size_type i = 0; i < raw.size(); ++i) {
        char c = raw[i];
        if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || (c >= '0' && c <= '9')) {
            if (needDash) {
                result += '-';
            }
            result += c;
            needDash = false;
        } else {
            needDash = true;
        }
    }
    return result;
}

/** Check for characters valid in identifiers (also see isValidIdentifier). */
static bool
isIdentifierCharacter(const char c)
{
    return (c >= '0' && c <= '9')
        || (c >= 'A' && c <= 'Z')
        || (c >= 'a' && c <= 'z')
        || c == '_';
}

/** Check for characters guaranteed not to be mangled by a shell. All other
    characters must be quoted somehow. */
static bool
isSafeCharacter(const char c)
{
    return isIdentifierCharacter(c)
        || c == '.'
        || c == '-'
        || c == '+'
        || c == '/'
        || c == ':'
        || c == ',';
}

/** Check for a valid identifier. Those are valid as variable names in a shell script,
    and thus can be possible exported values. */
static bool
isValidIdentifier(const string_t& s)
{
    /* bash(1) says:
       name   A word consisting only of  alphanumeric  characters  and  underscores,  and
              beginning  with an alphabetic character or an underscore.  Also referred to
              as an identifier.

       SUSv4 says in volume 1, 3.230 Name:
         In the shell command language, a word consisting solely of underscores, digits,
         and alphabetics from the portable character set. The first character of a name
         is not a digit. */
    if (s.empty() || (s[0] >= '0' && s[0] <= '9')) {
        return false;
    }
    for (string_t::size_type i = 0; i < s.size(); ++i) {
        if (!isIdentifierCharacter(s[i])) {
            return false;
        }
    }
    return true;
}

/** Export a value. Verifies that the key is valid, and escapes the value as needed.
    \param out output file (c2host.ini)
    \param key name
    \param value value */
static void
exportValue(std::ostream& out, const string_t key, const string_t value)
{
    if (!isValidIdentifier(key)) {
        // This value cannot be exported.
    } else {
        out << key << '=';
        for (string_t::size_type i = 0; i < value.size(); ++i) {
            if (uint8_t(value[i]) < ' ') {
                // embedded newline etc.
                break;
            } else if (!isSafeCharacter(value[i])) {
                out << '\\';
            }
            out << value[i];
        }
        out << '\n';
    }
}

/** Export a database hash into c2host.ini.
    \param out output file, c2host.ini
    \param prefix Prefix to use for all values
    \param has Hash to output */
static void
exportHash(std::ostream& out, const string_t prefix, DbHash hash)
{
    StringList_t data;
    hash.getAll(data);
    for (uint32_t i = 0; i+1 < data.size(); i += 2) {
        exportValue(out, prefix + "_" + data[i], data[i+1]);
    }
}

/** Export game environment. Creates a file c2host.ini, containing relevant
    config options in the form of environment variables, for use by the
    hosting scripts.
    \param dbConnection database connection, locked
    \param gameId game Id
    \return game directory */
string_t
exportEnvironment(DbClient& dbConnection, const int32_t gameId)
{
    Game game(DbSubtree(dbConnection, GAME_ROOT), gameId, Game::NoExistanceCheck);
    string_t gamedir = game.getDirectory();
    std::ofstream out((gamedir + "/c2host.ini").c_str());

    // Settings
    exportHash(out, "game_settings", game.getGame().hashKey("settings"));
    exportValue(out, "game_name", game.getName());

    // Host
    string_t host = game.getConfig("host");
    exportValue(out, "game_host", host);
    exportHash(out, "game_host", DbSubtree(dbConnection, HOST_PROGRAM_ROOT).subtree("prog").hashKey(host));

    // Master
    string_t master = game.getConfig("master");
    exportValue(out, "game_master", master);
    exportHash(out, "game_master", DbSubtree(dbConnection, MASTER_PROGRAM_ROOT).subtree("prog").hashKey(master));

    // Ship list
    string_t sl = game.getConfig("shiplist");
    exportValue(out, "game_sl", sl);
    exportHash(out, "game_sl", DbSubtree(dbConnection, SHIPLIST_PROGRAM_ROOT).subtree("prog").hashKey(sl));

    // Tools
    StringList_t tools;
    string_t toolList;
    game.getGame().hashKey("toolkind").getAll(tools);
    for (uint32_t i = 0; i+1 < tools.size(); i += 2) {
        exportValue(out, "game_tool_" + tools[i], tools[i+1]);
        exportHash(out, "game_tool_" + tools[i], DbSubtree(dbConnection, TOOL_PROGRAM_ROOT).subtree("prog").hashKey(tools[i+1]));
        if (i != 0) {
            toolList += " ";
        }
        toolList += tools[i];
    }
    exportValue(out, "game_tools", toolList);

    // Population
    string_t pop;
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (i != 1) {
            pop += ",";
        }
        if (game.isSlotInGame(i)) {
            pop += "yes";
        } else {
            pop += "no";
        }
    }
    exportValue(out, "game_slots", pop);

    return gamedir;
}

static void
respawnGame(DbClient& dbConnection, Game& game)
{
    DbSubtree root(dbConnection, GAME_ROOT);

    // Get the game this is a copy of
    int32_t copyOfId = game.getConfigInt("copyOf");
    if (!copyOfId) {
        return;
    }
    Game copyOf(root, copyOfId);

    // If this is a respawn cycle, get that Id
    int32_t newSourceId = copyOf.getConfigInt("copyNext");
    if (newSourceId == 0) {
        newSourceId = copyOfId;
    }
    Game newSource(root, newSourceId);

    // Finally, create the new game
    int32_t newId = createNewGame();
    copyGame(newSourceId, newId);
    createNewGameFinish(newId, "joining", game.getType());
    std::printf("[cron] [respawn] created game %d as copy of %d\n", int(newSourceId), int(newId));
}

static void
importGameData(DbClient& dbConnection, int32_t gameId)
{
    Game game(DbSubtree(dbConnection, GAME_ROOT), gameId, Game::NoExistanceCheck);
    string_t gamedir = game.getDirectory();

    // The game has hosted, so, obviously, master has run and the game is running
    game.setConfigInt("masterHasRun", 1);
    game.setState("running");
    game.clearCache();

    // Check pending copy
    if (game.getConfigInt("copyPending")) {
        // For copying a game, we need database access, which we have.
        // We're not going to modify games, so we do not need to lock them,
        // even if there is an ongoing transaction on them.
        game.setConfigInt("copyPending", 0);
        try {
            respawnGame(dbConnection, game);
        }
        catch (std::exception& e) {
            // This handler will be hit if the game refers to a non-existant game
            // as its copy-source.
            std::printf("[cron] [error] respawning this game failed, %s\n", e.what());
        }
    }

    // Read timestamp and turn number
    char nextturn_data[20];
    {
        Ptr<Stream> in = Stream::openFile(gamedir + "/data/nextturn.hst", Stream::C_READ);
        in->readT(nextturn_data, sizeof(nextturn_data));
    }
    string_t turnTime(nextturn_data, 18);
    int16_t turnNr(getInt16(&nextturn_data[18]));

    game.setConfigInt("turn", turnNr);
    game.setConfig("timestamp", turnTime);
    DbSubtree(dbConnection, GAME_ROOT).subtree("bytime").intKey(turnTime).set(gameId);

    // Times
    time_t now = getCurrentTime();
    game.setConfigInt("lastHostTime", now);
    game.setConfigInt("lastTurnSubmitted", 0);

    // Track history:
    // - players
    DbSubtree turn(game.getGame().subtree("turn").subtree(turnNr));
    for (int32_t i = 1; i <= NUM_PLAYERS; ++i) {
        turn.hashKey("player").stringField(format("%d", i)).set(game.getSlot(i).players()[0]);
    }

    // - scores
    importGameScores(gamedir + "/data", turn.hashKey("scores"), game.getGame().hashKey("scores"));

    // - turn status; also reset turn status
    char turnStatus[2*NUM_PLAYERS];
    string_t timScore = turn.hashKey("scores").getStringField("timscore");
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        DbHashIntField statusField = game.getSlot(i).turnStatus();
        int32_t slotStatus = game.getSlot(i).slotStatus().get();
        int32_t status = statusField.get();
        if (timScore.size() >= uint32_t(i)*4 && getInt32(&timScore[4*(i-1)]) == 0) {
            statusField.set(TurnDead);
        } else {
            statusField.set(TurnMissing);
        }
        storeInt16(&turnStatus[(i-1)*2], slotStatus ? status : -1);
    }
    turn.hashKey("info").setField("time", now);
    turn.hashKey("info").setField("timestamp", turnTime);
    turn.hashKey("info").setField("turnstatus", string_t(turnStatus, sizeof(turnStatus)));

    // - victory check
    if (checkVictory(gamedir + "/data", game)) {
        computeGameRankings(dbConnection, game);
        game.setState("finished");
    }

    // Reset filer
    ReconnectingGuard g(file_rc);
    FileClient file(file_connection);
    file.setRootPermissions();
    file.forgetDirectory(gamedir);
}

static int32_t
unimportGameData(DbClient& dbConnection, int32_t gameId)
{
    Game game(DbSubtree(dbConnection, GAME_ROOT), gameId, Game::NoExistanceCheck);

    // Remove bytime association (nobody will upload turns now anyway)
    string_t turnTime = game.getConfig("timestamp");
    DbSubtree(dbConnection, GAME_ROOT).subtree("bytime").intKey(turnTime).remove();

    return game.getConfigInt("turn");
}

static void
doRunHost(Runner& hostRunner, string_t gameDir, int32_t turn)
{
    // Wait before running host to avoid generating duplicate timestamps
    sleep(2);

    // Build command line
    printf("[cron] %s: running host\n", gameDir.c_str());
    IntDataSegment cmdl;
    cmdl.pushNew(makeStringValue("/bin/sh"));
    cmdl.pushNew(makeStringValue("bin/runhost.sh"));
    cmdl.pushNew(makeStringValue(gameDir));
    cmdl.pushNew(makeStringValue(format("%03d", turn)));

    // Run it
    string_t output;
    int32_t error = hostRunner.run(cmdl, output);
    if (error != 0) {
        string_t err = format("Host failed with error %d. Output follows:\n", error, output);
        throw std::runtime_error(err.c_str());
    }
}

/** Find next host date. Must be called after importGameData().
    This is a heavily stripped-down version of doScheduleShow
    to return precisely one schedule. */
static uint32_t
getNextHostDate(Game& game, bool& hostEarly, Schedule::Type& type)
{
    // Figure out last host date
    int32_t lastHost = game.getConfigInt("lastHostTime");
    int32_t turn     = game.getConfigInt("turn");

    // Process all schedules
    DbSubtree sroot(game.getSchedule());
    DbStringList schedules(sroot.stringListKey("list"));
    int32_t numSchedules = schedules.size();
    int32_t currentSchedule = 0;
    while (currentSchedule < numSchedules) {
        // Process one schedule
        Schedule sched;
        sched.loadFrom(sroot.hashKey(schedules[currentSchedule]));
        if (!sched.isExpired(turn, lastHost)) {
            hostEarly = sched.getHostEarly();
            type = sched.getType();
            return sched.getNextHost(lastHost);
        }
        ++currentSchedule;
    }
    hostEarly = false;
    type = Schedule::Stopped;
    return 0;
}

static void
describeNextHostDate(MailClient& mail_connection,
                     uint32_t nextHost,
                     bool nextHostEarly,
                     Schedule::Type nextHostType)
{
    string_t suffix = nextHostEarly ? "_early" : "";
    if (nextHost != 0) {
        char buffer[100];
        time_t now = getSystemTimeFromTime(nextHost);
        strftime(buffer, sizeof(buffer), "%d/%b/%Y %H:%M GMT", gmtime(&now));
        mail_connection.addParameter("next_schedule", "day" + suffix);
        mail_connection.addParameter("next_day_time", buffer);
    } else {
        if (nextHostType == Schedule::Manual) {
            mail_connection.addParameter("next_schedule", "manual" + suffix);
        } else if (nextHostType == Schedule::Stopped) {
            mail_connection.addParameter("next_schedule", "stop" + suffix);
        } else {
            mail_connection.addParameter("next_schedule", "quick" + suffix);
        }
    }
}

enum {
    Result,
    ResultPlayerFiles,
    Zip,
    ZipPlayerFiles,
    Info,
    NumFormats
};

static void
collectPlayers(Game& game, StringList_t& players, DbSubtree& user,
               StringList_t (&playersByFormat)[NumFormats])
{
    // Get mail formats.
    for (uint32_t i = 0; i < players.size(); ++i) {
        string_t fmt = game.getPlayerConfig(players[i], "mailgametype");
        bool hasPF = game.getPlayerConfigInt(players[i], "hasPlayerFiles");
        if (!hasPF) {
            game.setPlayerConfigInt(players[i], "hasPlayerFiles", true);
        }
        if (fmt.empty() || fmt == "default") {
            // FIXME: this implements srvProfileGet() by hand. Can we do better?
            fmt = user.subtree(players[i]).hashKey("profile").getStringField("mailgametype");
            if (fmt.empty()) {
                fmt = DbHash(user.getClient(), "default:profile").getStringField("mailgametype");
            }
        }
        if (fmt == "none") {
            // User does not want mail
        } else if (fmt == "rst") {
            playersByFormat[Result + !hasPF].push_back("user:" + players[i]);
        } else if (fmt == "info") {
            playersByFormat[Info].push_back("user:" + players[i]);
        } else {
            playersByFormat[Zip + !hasPF].push_back("user:" + players[i]);
        }
    }
}

struct ResultMailInfo {
    int32_t gameId;
    string_t gameName;
    string_t gameUrl;
    string_t gameDir;
    int32_t gameTurn;
    bool nextHostEarly;
    bool finalTurn;
    Schedule::Type nextHostType;
    uint32_t nextHostTime;

    void init(Game& game);
    void describeGame(MailClient& mailer);
    void sendResults(MailClient& mailer, Game& game, int slot, StringList_t (&playersByFormat)[NumFormats]);
};

void
ResultMailInfo::init(Game& game)
{
    gameId       = game.getId();
    gameName     = game.getName();
    gameUrl      = makeGameUrl(gameId, gameName);
    gameDir      = game.getDirectory();
    gameTurn     = game.getConfigInt("turn");
    nextHostTime = getNextHostDate(game, nextHostEarly, nextHostType);
    finalTurn    = game.getState() == "finished";
}

void
ResultMailInfo::describeGame(MailClient& mailer)
{
    mailer.addParameter("gameid", gameId);
    mailer.addParameter("gamename", gameName);
    mailer.addParameter("gameurl", gameUrl);
    mailer.addParameter("gameturn", gameTurn);
    describeNextHostDate(mailer, nextHostTime, nextHostEarly, nextHostType);
}

void
ResultMailInfo::sendResults(MailClient& mailer, Game& game, int slot,
                            StringList_t (&playersByFormat)[NumFormats])
{
    static const char*const suffixes[] = {
        "-rst",
        "-rst-pf",
        "",
        "-pf",
        "-info",
    };

    for (int fmt = 0; fmt < NumFormats; ++fmt) {
        string_t id = format("result-%d-%d%s", gameId, slot, suffixes[fmt]);
        if (playersByFormat[fmt].empty()) {
            // Nobody wants this format, just cancel the previous mail.
            mailer.cancel(id);
        } else {
            // Someone wants this format, so generate the message
            string_t tpl;
            if (finalTurn) {
                tpl += "last-";
            }
            if (fmt == Info) {
                tpl += "result-info";
            } else {
                tpl += "result";
            }
            mailer.startMail(tpl, id);
            mailer.addParameter("slot", slot);
            describeGame(mailer);

            if (finalTurn) {
                // A slot is marked dead only when the last player resigns.
                // That is, if we actually have a player here, the slot is alive,
                // and the slot will have a nonzero rank assigned.
                mailer.addParameter("rank", game.getSlot(slot).rank().get());
            }

            if (fmt == Zip || fmt == ZipPlayerFiles) {
                mailer.attachFile(format("c2file://%s:%s/%s/out/%d/player%d.zip")
                                  << file_rc.getHost()
                                  << file_rc.getPort()
                                  << gameDir << slot << slot);
            }
            if (fmt == Result || fmt == ResultPlayerFiles) {
                // xref runhost.sh
                static const char* const files[] = {
                    "player%d.rst",
                    "util%d.dat",
                    "xyplan%d.dat",
                };
                for (uint_t i = 0; i < countof(files); ++i) {
                    string_t baseName = format(files[i], slot);
                    string_t name = format("%s/out/%d/%s", gameDir, slot, baseName);
                    if (Stream::openFileNT(name, Stream::C_READ)) {
                        mailer.attachFile(format("c2file://%s:%s/%s",
                                                 file_rc.getHost(),
                                                 file_rc.getPort(),
                                                 name));
                    }
                }
            }
            if (fmt == ZipPlayerFiles || fmt == ResultPlayerFiles) {
                mailer.attachFile(format("c2file://%s:%s/%s/out/all/playerfiles.zip")
                                  << file_rc.getHost()
                                  << file_rc.getPort()
                                  << gameDir);
            }

            // Send it
            mailer.sendTo(playersByFormat[fmt]);
        }
    }
}

static void
sendResultFiles(DbClient& dbConnection, int32_t gameId)
{
    Game game(DbSubtree(dbConnection, GAME_ROOT), gameId, Game::NoExistanceCheck);
    DbSubtree user(dbConnection, USER_ROOT);

    ResultMailInfo rmi;
    rmi.init(game);

    std::map<string_t, GPlayerSet> allPlayers;
    for (int slot = 1; slot <= NUM_PLAYERS; ++slot) {
        // List players
        StringList_t playersByFormat[NumFormats];
        StringList_t players;
        game.listPlayers(slot, players);
        for (uint_t i = 0; i < players.size(); ++i) {
            allPlayers[players[i]] += slot;
        }

        // Send results in all formats
        ReconnectingGuard g(mail_rc);
        collectPlayers(game, players, user, playersByFormat);
        rmi.sendResults(mail_connection, game, slot, playersByFormat);
    }

    // Distribute results to local directories
    for (std::map<string_t, GPlayerSet>::iterator i = allPlayers.begin(); i != allPlayers.end(); ++i) {
        string_t userGameDir = game.getPlayerConfig(i->first, "gameDir");
        if (!userGameDir.empty()) {
            ReconnectingGuard g(userfile_rc);
            FileClient file(userfile_connection);
            file.setUserPermissions(i->first);
            installGameData(game, i->second, file, userGameDir);
        }
    }
}

static void
importMissingTurns(DbClient& dbConnection, ReconnectingConnection& dbLock, string_t gameDir, int32_t gameId, int32_t slot)
{
    // Look for players
    Game game(DbSubtree(dbConnection, GAME_ROOT), gameId, Game::NoExistanceCheck);
    Game::Slot gameSlot(game.getSlot(slot));
    StringList_t players;
    {
        ReconnectingGuard g(dbLock);
        int32_t existingState = gameSlot.turnStatus().get() & TurnStateMask;
        if (existingState == TurnYellow || existingState == TurnGreen) {
            // Turn is there
            return;
        }
        game.listPlayers(slot, players);
    }

    // When we're here, there is no turn, but we may have some players
    for (uint32_t i = 0; i < players.size(); ++i) {
        ReconnectingGuard g(dbLock);

        // Check whether user has a game directory configured
        // FIXME: should work backwards! current player is last on list!
        string_t userGameDir = game.getPlayerConfig(players[i], "gameDir");
        if (userGameDir.empty()) {
            continue;
        }

        // Check whether user has a turn file
        try {
            // Load from userfile
            string_t trnData;
            {
                router_client.tryCall(format("CLOSE -WDIR=%s\n", userGameDir));
                ReconnectingGuard gg(userfile_rc);
                FileClient file(userfile_connection);
                file.setUserPermissions(players[i]);
                trnData = file.getFile(format("%s/player%d.trn", userGameDir, slot));;
            }
            printf("[cron] %s: trying turn from 'user:%s'\n", gameDir.c_str(), userGameDir.c_str());

            // Store on hostfile
            {
                ReconnectingGuard gg(file_rc);
                FileClient file(file_connection);
                file.setUserPermissions(players[i]);
                file.putFile(format("%s/in/new/player%d.trn", gameDir, slot), trnData);
            }

            // Run checkturn
            IntDataSegment cmd;
            cmd.pushNew(makeStringValue("/bin/sh"));
            cmd.pushNew(makeStringValue("bin/checkturn.sh"));
            cmd.pushNew(makeStringValue(gameDir));
            cmd.pushNew(makeIntValue(slot));
            string_t output;
            int32_t code = checkturn_runner.run(cmd, output);
            int32_t newState = (code + 1);
            if (newState == TurnGreen || newState == TurnYellow) {
                // Success!
                printf("[cron] %s: turn file from 'user:%s' succeeded\n", gameDir.c_str(), userGameDir.c_str());
                gameSlot.turnStatus().set(newState);

                // Do not update replacements. We're currently sweeping together fragments
                // of a partially played turn to minimize damages, so don't damage things
                // even more by changing subscriptions. Plus, it'd be quite a nightmare to
                // start messing around with subscriptions at this point, which would also
                // update game directories beneath our back.
                break;
            }
        }
        catch (...) {
            // Ignore any errors
        }
    }
}

static void
processTurnStatus(DbClient& dbConnection, int32_t gameId)
{
    Game game(DbSubtree(dbConnection, GAME_ROOT), gameId, Game::NoExistanceCheck);

    // No ranking if it is disabled or this is a test game
    if (game.isRankingDisabled()) {
        return;
    }

    std::map<string_t, uint32_t> playerValues;
    const uint32_t nil = uint32_t(-1);

    // Process all slots
    for (int slot = 1; slot <= NUM_PLAYERS; ++slot) {
        // Get status and players
        int32_t state = game.getSlot(slot).turnStatus().get() & TurnStateMask;
        StringList_t players;
        game.listPlayers(slot, players);

        // Collect player values. Backwards, current player is last of list!
        for (uint32_t i = 0; i < players.size(); ++i) {
            const string_t& thisPlayer = players[players.size()-1-i];
            const uint32_t thisPlayerValue = (state == TurnGreen || state == TurnYellow || state == TurnDead) ? nil : i;

            // Player gets maximum possible penalty earned this turn. Maximum possible
            // penalty is thisPlayerValue=0 (missed a turn as primary player), minimum
            // is nil (submitted a turn). Thus we're looking for the minimum thisPlayerValue.
            std::map<string_t, uint32_t>::iterator it = playerValues.find(thisPlayer);
            if (it != playerValues.end()) {
                it->second = std::min(it->second, thisPlayerValue);
            } else {
                playerValues.insert(std::make_pair(thisPlayer, thisPlayerValue));
            }
        }
    }

    // Process it
    for (std::map<string_t, uint32_t>::const_iterator it = playerValues.begin(); it != playerValues.end(); ++it) {
        if (it->second != nil) {
            // Miss
            handlePlayerTurn(dbConnection, it->first, false, it->second);
        } else {
            // Submit
            handlePlayerTurn(dbConnection, it->first, true, 0);
        }
        handlePlayerRankChanges(dbConnection, it->first);
    }
}

/*************************** External Interface **************************/

void
runHost(Runner& hostRunner, DbClient& dbConnection, ReconnectingConnection& dbLock, int32_t gameId)
{
    // Export environment
    string_t gameDir;
    int32_t turnNr;
    {
        ReconnectingGuard g(dbLock);
        turnNr = unimportGameData(dbConnection, gameId);
        gameDir = exportEnvironment(dbConnection, gameId);
    }

    // Locate missing turns
    for (int32_t i = 1; i <= NUM_PLAYERS; ++i) {
        importMissingTurns(dbConnection, dbLock, gameDir, gameId, i);
    }

    // Run host
    doRunHost(hostRunner, gameDir, turnNr+1);

    // Re-import VGAP data
    {
        ReconnectingGuard g(dbLock);
        processTurnStatus(dbConnection, gameId);
        importGameData(dbConnection, gameId);
        sendResultFiles(dbConnection, gameId);
    }

    printf("[cron] %s: host completed\n", gameDir.c_str());
}

void
runMaster(Runner& hostRunner, DbClient& dbConnection, ReconnectingConnection& dbLock, int32_t gameId)
{
    // Export environment
    string_t gameDir;
    {
        ReconnectingGuard g(dbLock);
        gameDir = exportEnvironment(dbConnection, gameId);
    }

    // Build command line
    printf("[cron] %s: running master\n", gameDir.c_str());
    IntDataSegment cmdl;
    cmdl.pushNew(makeStringValue("/bin/sh"));
    cmdl.pushNew(makeStringValue("bin/runmaster.sh"));
    cmdl.pushNew(makeStringValue(gameDir));

    // Run it
    {
        string_t output;
        int32_t error = hostRunner.run(cmdl, output);
        if (error != 0) {
            string_t err = format("Master failed with error %d. Output follows:\n%s", error, output);
            throw std::runtime_error(err.c_str());
        }
    }

    // OK, proceed with running host
    doRunHost(hostRunner, gameDir, 1);

    // Re-import VGAP data
    {
        ReconnectingGuard g(dbLock);
        importGameData(dbConnection, gameId);
        sendResultFiles(dbConnection, gameId);
    }

    printf("[cron] %s: master completed\n", gameDir.c_str());
}
