/**
  *  \file planetscentral/host/newgame.cc
  *  \brief Host: Game creation
  */

#include "planetscentral/host/newgame.h"
#include "cpluslib/format.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/host/host.h"
#include "planetscentral/host/db.h"
#include "planetscentral/dbclient/fileclient.h"
#include "game/struct.h"
#include "planetscentral/host/cron.h"
#include "planetscentral/host/schedule.h"
#include "cpluslib/misc.h"
#include "planetscentral/host/game.h"

/** Create a new game. This will automatically obtain DB/filer access, so you
    need not have it locked.
    \return game Id */
int32_t
createNewGame()
{
    DbSubtree root(database_connection, GAME_ROOT);

    int32_t id;
    string_t dirName;

    // Database operations
    {
        ReconnectingGuard g(database_rc);

        // Allocate new Id
        id = ++root.intKey("lastid");
        dirName = format("games/%04d", id);

        DbSubtree game(root.subtree(id));
        game.stringKey("dir").set(dirName);
    }

    // File operations
    {
        ReconnectingGuard g(file_rc);
        FileClient file(file_connection);
        file.setRootPermissions();
        try {
            file.removeDirectory(dirName);
        }
        catch (...)
        { }
        file.createDirectory(dirName);
        file.createDirectory(dirName + "/data");
        file.createDirectory(dirName + "/backup");
        file.createDirectory(dirName + "/in");
        file.createDirectory(dirName + "/in/new");
        file.createDirectory(dirName + "/out");
        file.createDirectory(dirName + "/out/all");
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            file.createDirectory(format("%s/out/%d", dirName, i));
        }
    }

    return id;
}

/** Finish game creation. This places the game on the respective lists
    so it can be found by other commands.
    \param id Game id
    \param state Game state (running, preparing, joining)
    \param type Game type (public, private, test, unlisted) */
void
createNewGameFinish(const int32_t id, const string_t state, const string_t type)
{
    const char PREPARE_STATE[] = "preparing";
    ReconnectingGuard g(database_rc);
    DbSubtree root(database_connection, GAME_ROOT);
    DbSubtree game(root.subtree(id));

    // Create the game in state "preparing". This is less efficient than creating it
    // in the correct state directly, but allows us to re-use the transition handling
    // of Game::setState for proper interaction with the rest of the system.
    game.stringKey("state").set(PREPARE_STATE);
    game.stringKey("type").set(type);
    root.subtree("state").intSetKey(PREPARE_STATE).add(id);
    if (type == "public") {
        root.subtree("pubstate").intSetKey(PREPARE_STATE).add(id);
    }
    root.intSetKey("all").add(id);

    // Perform a regular state transition
    Game(root, id, Game::NoExistanceCheck).setState(state);
}

/********************************** Copy *********************************/

/** Copy a game's tools.
    \param src Source game
    \param dst Destination game */
static void
copyTools(DbSubtree& src, DbSubtree& dst)
{
    StringList_t tools;
    src.stringSetKey("tools").getAll(tools);
    for (std::size_t i = 0; i < tools.size(); ++i) {
        // - tools
        const string_t& tool = tools[i];
        dst.stringSetKey("tools").add(tool);

        // - tool:$TOOL:settings
        StringList_t settings;
        src.subtree("tool").subtree(tool).hashKey("settings").getAll(settings);
        if (!settings.empty()) {
            dst.subtree("tool").subtree(tool).hashKey("settings").setAll(settings);
        }

        // - toolkind. Copy from master data, not from source game.
        string_t kind = DbSubtree(database_connection, TOOL_PROGRAM_ROOT).subtree("prog").hashKey(tool).getStringField("kind");
        if (!kind.empty()) {
            dst.hashKey("toolkind").setField(kind, tool);
        }
    }
}

/** Copy a game's schedules.
    \param src Source game
    \param dst Destination game
    \param daytime Place destination schedules on this daytime */
static void
copySchedule(DbSubtree& src, DbSubtree& dst, int32_t daytime)
{
    DbSubtree srcSched(src.subtree("schedule"));
    DbSubtree dstSched(dst.subtree("schedule"));

    // Get list of schedules
    IntList_t list;
    srcSched.intListKey("list").getAll(list);

    // Copy them one by one, modifying the daytime
    for (std::size_t i = list.size(); i > 0; --i) {
        Schedule sch;
        sch.loadFrom(srcSched.hashKey(format("%d", list[i-1])));
        sch.setDaytime(daytime);

        int32_t newId = ++dstSched.intKey("lastId");
        sch.saveTo(dstSched.hashKey(format("%d", newId)));
        dstSched.intListKey("list").pushFront(newId);
    }
}

/** Copy a game. For best results, the source game should be preparing, but other
    states are handled as well. The destination game should be new. */
void
copyGame(int32_t srcId, int32_t dstId)
{
    ReconnectingGuard g(database_rc);

    DbSubtree root(database_connection, GAME_ROOT);
    DbSubtree src(root.subtree(srcId));
    DbSubtree dst(root.subtree(dstId));

    // Build new name
    string_t srcName = src.stringKey("name").get();
    dst.stringKey("name").set(format("%s %d", srcName, ++root.hashKey("bynameprefix").intField(srcName)));

    // Copy description
    dst.stringKey("description").set(src.stringKey("description").get());

    // Open slots
    for (int slot = 1; slot <= NUM_PLAYERS; ++slot) {
        // FIXME: if this game has run, use game:$GID:turn:1:info->turnstatus to find open slots
        DbHash dstH = dst.subtree("player").subtree(slot).hashKey("status");
        DbHash srcH = src.subtree("player").subtree(slot).hashKey("status");
        dstH.setField("slot", srcH.getIntField("slot"));
        dstH.setField("turn", 0);
    }

    // Copy schedule
    copySchedule(src, dst, pickDayTime(root));

    // Copy settings
    DbHash srcSet = src.hashKey("settings");
    DbHash dstSet = dst.hashKey("settings");

    // - copy some fields raw
    static const char*const fieldsToCopy[] = {
        // - do not copy lastHostTime, lastTurnSubmitted, lastPlayerJoined, nextHostTime
        // - do not copy turn, timestamp, rankTurn, hostRunNow
        // - do not copy copyEnable, copyNext
        "host",
        "master",
        "shiplist",
        "endCondition",
        "endTurn",
        "endProbability",
        "endScore",
        "endScoreName",
        "rankDisable",
    };
    for (std::size_t i = 0; i < countof(fieldsToCopy); ++i) {
        std::auto_ptr<IntValue> v(srcSet.getFieldRaw(fieldsToCopy[i]));
        if (v.get() != 0) {
            dstSet.setFieldRaw(fieldsToCopy[i], v.get());
        }
    }

    // - set some fields to fixed values
    dstSet.setField("masterHasRun", 0);
    dstSet.setField("copyOf", srcId);

    // Copy tools
    copyTools(src, dst);

    // Do not copy state. This is set by createNewGameFinish.
    // Do not copy type. This is set by createNewGameFinish.
    // Do not copy owner.
}
