/**
  *  \file planetscentral/host/talk.cc
  *  \brief Interface to Forum
  */

#include <cctype>
#include "planetscentral/host/talk.h"
#include "planetscentral/host/game.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/host/host.h"
#include "cpluslib/format.h"

/** Simplify a topic name for inclusion in an URL. */
static string_t
simplifyNewsgroupName(string_t s)
{
    string_t result;
    bool needsp = false;
    for (string_t::size_type i = 0; i < s.size(); ++i) {
        char ch = s[i];
        if ((ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z') || (ch >= '0' && ch <= '9')) {
            if (needsp) {
                result += '-';
            }
            result += std::tolower(ch);
            needsp = false;
        } else {
            needsp = (result.size() != 0);
        }
    }
    return result;
}

static void
setParent(DbRequest& req, bool isPublic)
{
    req.withString("parent").withString(isPublic ? "active" : "active-unlisted");
}

static void
setName(DbRequest& req, const string_t& gameName)
{
    req.withString("name").withString(gameName);
    req.withString("key").withString(strLCase(gameName));
}

static void
setPermissions(DbRequest& req, bool isPublic, int32_t gameId)
{
    if (isPublic) {
        req.withString("readperm").withString("all");
        req.withString("writeperm").withString("-u:anon,p:allowpost");
        req.withString("answerperm").withString("-u:anon,p:allowpost");
    } else {
        string_t perm = format("g:%d", gameId);
        req.withString("readperm").withString(perm);
        req.withString("writeperm").withString(perm);
        req.withString("answerperm").withString(perm);
    }
}

/** Handle game start. Invoked when a game goes into state "running" or "joining".
    This is where the forum is created (and nowhere else).

    Implementation:
    - if it's public, make a public forum
    - if it's private or unlisted, make an unlisted, private forum
    - otherwise, don't make a forum */
void
talkHandleGameStart(Game& g, string_t gameType)
{
    bool isPublic = (gameType == "public");
    bool isPrivate = (gameType == "private" || gameType == "unlisted");
    if ((!isPublic && !isPrivate) || g.getConfigInt("forum") != 0 || g.getConfigInt("forumDisable") != 0) {
        // It is neither public nor private, or already has a forum, or should not have one.
        return;
    }

    // Set up forum
    string_t gameName = g.getName();
    ReconnectingGuard guard(talk_rc);
    DbRequest req;
    req.withString("FORUMADD");
    setParent(req, isPublic);
    setPermissions(req, isPublic, g.getId());
    setName(req, gameName);
    req.withString("description").withString(format("forum:Forum for [game]%d[/game] (#%0$d)", g.getId()));
    req.withString("newsgroup").withString("planetscentral.games." + simplifyNewsgroupName(format("%d-%s", g.getId(), simplifyNewsgroupName(gameName))));

    // Create it
    int32_t forumId = talk_connection.callInt(req);
    g.setConfigInt("forum", forumId);
}

/** Handle game end. Invoked when a game goes into state "finished".

    Implementation: if the game has a forum which is in an "active" category,
    move it into the corresponding "finished" category. If the forum has already
    been manually moved, does nothing. */
void
talkHandleGameEnd(Game& g, string_t /*gameType*/)
{
    // Get forum Id
    int32_t forumId = g.getConfigInt("forum");
    if (forumId == 0) {
        return;
    }

    // Update forum
    ReconnectingGuard guard(talk_rc);
    string_t parent = talk_connection.callString(DbRequest().withString("FORUMGET").withInt(forumId).withString("parent"));
    if (parent.size() >= 6 && parent.compare(0, 6, "active", 6) == 0) {
        parent.replace(0, 6, "finished", 8);
        talk_connection.callString(DbRequest().withString("FORUMSET").withInt(forumId).withString("parent").withString(parent));
    }
}

/** Handle game name change. Called whenever a game name changes. This will just update
    the forum name and key. */
void
talkHandleGameNameChange(Game& g, string_t newName)
{
    // Get forum Id
    int32_t forumId = g.getConfigInt("forum");
    if (forumId == 0) {
        return;
    }

    // Update forum
    ReconnectingGuard guard(talk_rc);
    DbRequest req;
    req.withString("FORUMSET").withInt(forumId);
    setName(req, newName);
    talk_connection.callString(req);
}

/** Handle game type change. Called when a game's type changes.

    Implementation: if the game has a forum, and it is in an "active" category,
    move it to its proper "active" category. If the forum already is in another
    category (by having been moved manually, or by being in a "finished" category,
    this does not change the forum. */
void
talkHandleGameTypeChange(Game& g, string_t /*gameState*/, string_t gameType)
{
    // Get forum Id
    int32_t forumId = g.getConfigInt("forum");
    if (forumId == 0) {
        return;
    }

    // Check existing parent. Only if it's active, move it.
    string_t parent = talk_connection.callString(DbRequest().withString("FORUMGET").withInt(forumId).withString("parent"));
    if (parent.size() >= 6 && parent.compare(0, 6, "active", 6) == 0) {
        // Update forum
        bool isPublic = (gameType == "public");

        ReconnectingGuard guard(talk_rc);
        DbRequest req;
        req.withString("FORUMSET").withInt(forumId);
        setParent(req, isPublic);
        setPermissions(req, isPublic, g.getId());
        talk_connection.callString(req);
    }
}
