/**
  *  \file planetscentral/net/mimeparser.h
  *  \brief MIME Parsing
  */
#ifndef C2SERVER_PLANETSCENTRAL_NET_MIMEPARSER_H
#define C2SERVER_PLANETSCENTRAL_NET_MIMEPARSER_H

#include <cstddef>
#include <map>
#include <vector>
#include "cpluslib/string.h"

/** MIME Message Parsing. This splits a message into header and body,
    unfolds and decodes header fields, and decodes the body.

    Header fields are decoded into UTF-8. They are placed in a map,
    indexed by lowercase header field name. Note that this means that
    we can only store one instance of a repeated header field
    ("Received"). MIME "encoded-words" are decoded for all header
    fields. This is probably not perfect.

    The body is decoded from its transfer encoding. It is stored as a
    series of lines. There is an implicit newline BETWEEN body lines.
    That is, the empty body is represented by a one-element array with
    an empty string; a two-element array represents the body
    'a[0] + "\n" + a[1]'.

    Carriage-return characters are ignored. */
class MimeParser {
 public:
    typedef std::map<string_t, string_t> HeaderMap_t;
    typedef std::vector<string_t> BodyVec_t;

    MimeParser();

    // Manipulation/Setup
    void addData(const char* data, std::size_t size);
    void finish();
    void clear();

    // Access
    const HeaderMap_t& getHeaders() const;
    const BodyVec_t& getBody() const;
    string_t getHeader(string_t name) const;
    string_t getFileName() const;
    string_t getBodyAsString() const;

 private:
    std::size_t addDataHeader(const char* data, std::size_t size);
    std::size_t addDataPlainBody(const char* data, std::size_t size);
    std::size_t addDataBase64Body(const char* data, std::size_t size);
    std::size_t addDataQPBody(const char* data, std::size_t size);
    void flushBase64Body();

    /** Headers of mail. Indexed by lowercase header field name. */
    std::map<string_t, string_t> m_Headers;

    /** Body of mail, one entry per line. */
    std::vector<string_t> m_Body;

    /** Accumulator for parsing. */
    string_t m_Accum;

    /** Parser state. */
    enum {
        StateHeader,
        StatePlainBody,
        StateBase64Body,
        StateQPBody
    } m_State;

    /** StateHeader: Pointer to last header. */
    string_t* m_pLastHeader;

    void processHeader();
};

/** MIME Multipart Parser. This takes a MimeParser containing a multipart
    message and extracts individual parts into new MimeParser instances. */
class MimeMultipartParser {
 public:
    MimeMultipartParser();
    bool init(MimeParser& p);
    bool readPart(MimeParser& part);

 private:
    bool findNextBoundary();

    MimeParser* m_pParser;

    string_t m_Boundary;
    MimeParser::BodyVec_t::size_type m_Position;
};

string_t getHeaderPrimaryValue(string_t v);
string_t getHeaderSubValue(string_t v, string_t key);
string_t getHeaderAddressValue(string_t v);


/******************************** Inlines ********************************/

/** Get headers. */
inline const MimeParser::HeaderMap_t&
MimeParser::getHeaders() const
{
    return m_Headers;
}

/** Get body. */
inline const MimeParser::BodyVec_t&
MimeParser::getBody() const
{
    return m_Body;
}

#endif
