/**
  *  \file planetscentral/talk/htmlout.cc
  *  \brief Talk: HTML output
  */

#include <cctype>
#include "planetscentral/talk/htmlout.h"
#include "cpluslib/types.h"
#include "planetscentral/talk/text.h"
#include "util/misc.h"
#include "planetscentral/dbclient/dbsubtree.h"
#include "planetscentral/talk/db.h"
#include "planetscentral/talk/render.h"
#include "planetscentral/talk/user.h"
#include "planetscentral/talk/url.h"
#include "planetscentral/talk/forum.h"
#include "cpluslib/misc.h"
#include "cpluslib/utf8.h"
#include "planetscentral/talk/message.h"
#include "planetscentral/talk/topic.h"
#include "cpluslib/format.h"

namespace {
    class HtmlRenderer {
     public:
        HtmlRenderer(RenderState& state, DbClient& db, string_t& result)
            : state(state), db(db), result(result)
            { }
        void renderText(const string_t& text);
        void renderAttribution(const string_t& text);
        bool renderLink(TextNode* n, const char* pre, const char* mid, const char* post);
        bool renderGameLink(TextNode* n);
        bool renderForumLink(TextNode* n);
        bool renderUserLink(TextNode* n);
        bool renderPostLink(TextNode* n);
        bool renderThreadLink(TextNode* n);
        bool renderImage(TextNode* n);
        void renderChildren(TextNode* n);
        void render(TextNode* n);

     private:
        RenderState& state;
        DbClient& db;
        string_t& result;
    };
}

/** Check for valid URLs.
    The main reason for this function is to avoid users executing Javascript
    in other users' context. */
static bool
isValidUrl(string_t s)
{
    const char*const permittedPrefixes[] = {
        "/",
        "http://",
        "https://",
        "mailto:",
        "ftp://",
        "news:",
        "nntp:",
        "data:image/",
        "data:text/plain",
        "data:text/html"
    };
    for (std::size_t i = 0; i < countof(permittedPrefixes); ++i) {
        std::size_t n = std::strlen(permittedPrefixes[i]);
        if (s.size() >= n && s.compare(0, n, permittedPrefixes[i], n) == 0) {
            return true;
        }
    }
    return false;
}

static string_t
abbreviate(string_t s)
{
    if (s.size() > 30) {
        std::size_t i = 28;
        while (i > 0 && (s[i-1] == ' ' || Utf8Reader::isContinuationByte(s[i]))) {
            // Yes, this is intended to check i-1 for spaces, and i for continuation
            --i;
        }
        s.erase(i);
        s += "...";
    }
    return s;
}

static const char*
getInlineTagName(uint8_t minor)
{
    switch (minor) {
     case miInBold:             return "b";
     case miInItalic:           return "em";
     case miInStrikeThrough:    return "s";
     case miInUnderline:        return "u";
     case miInMonospace:        return "tt";
     default:                   return 0;
    }
}

void
HtmlRenderer::renderText(const string_t& text)
{
    string_t::size_type i = 0, n;
    while ((n = text.find_first_of("<>&\"'", i)) != text.npos) {
        result.append(text, i, n-i);
        switch (text[n]) {
         case '<':  result += "&lt;";   break;
         case '>':  result += "&gt;";   break;
         case '&':  result += "&amp;";  break;
         case '"':  result += "&quot;"; break;
         case '\'': result += "&apos;"; break;
        }
        i = n+1;
    }
    result.append(text, i, text.size()-i);
}

void
HtmlRenderer::renderAttribution(const string_t& text)
{
    string_t::size_type i = text.rfind(';');
    if (i != text.npos && i != text.size()-1) {
        // It is the form "user;posting"
        TextNode n;
        n.major = maLink;
        n.minor = miLinkUser;
        n.text.assign(text, 0, i);
        if (!renderUserLink(&n)) {
            renderText(n.text);
        }
        renderText(" in ");
        n.minor = miLinkPost;
        n.text.assign(text, i+1, string_t::npos);
        if (!renderPostLink(&n)) {
            renderText(n.text);
        }
    } else {
        // Just a user name
        TextNode n;
        n.major = maLink;
        n.minor = miLinkUser;
        n.text  = text;
        if (!renderUserLink(&n)) {
            renderText(n.text);
        }
    }
    renderText(":");
}

bool
HtmlRenderer::renderLink(TextNode* n, const char* pre, const char* mid, const char* post)
{
    result += pre;
    renderText(n->text);
    result += mid;
    if (n->children.empty()) {
        renderText(n->text);
    } else {
        renderChildren(n);
    }
    result += post;
    return true;
}

bool
HtmlRenderer::renderGameLink(TextNode* n)
{
    // Parse game Id
    int32_t gameId;
    if (!stringValue(n->text, gameId) || gameId <= 0) {
        return false;
    }

    // Access game, check permissions
    // xref host/game.cc, Game::hasPermission
    DbSubtree root(db, GAME_ROOT);
    if (!root.intSetKey("all").contains(gameId)) {
        return false;
    }

    DbSubtree game(root.subtree(gameId));
    const string_t gameState = game.stringKey("state").get();
    if (gameState != "joining" && gameState != "running" && gameState != "finished") {
        return false;
    }

    const string_t gameType = game.stringKey("type").get();
    if (gameType != "unlisted"
        && gameType != "public"
        && game.stringKey("owner").get() != state.getUser()
        && !game.hashKey("users").hasField(state.getUser()))
    {
        return false;
    }

    // OK, we are allowed to access it. Get its name.
    const string_t name = game.stringKey("name").get();
    result += "<a href=\"";
    renderText(state.getBaseUrl());
    result += makeGameUrl(gameId, name);
    result += "\">";
    if (n->children.empty()) {
        renderText(name);
    } else {
        renderChildren(n);
    }
    result += "</a>";
    return true;
}

bool
HtmlRenderer::renderForumLink(TextNode* n)
{
    // Parse forum Id
    int32_t forumId;
    if (!stringValue(n->text, forumId) || forumId <= 0) {
        return false;
    }

    // Access forum, check permissions
    Forum forum(db, forumId);
    if (!forum.exists()) {
        // FIXME permission check!
        return false;
    }

    // OK, we are allowed to access it. Get its name.
    const string_t name = forum.name().get();
    result += "<a href=\"";
    renderText(state.getBaseUrl());
    result += makeForumUrl(forumId, name);
    result += "\">";
    if (n->children.empty()) {
        renderText(name);
    } else {
        renderChildren(n);
    }
    result += "</a>";
    return true;
}

bool
HtmlRenderer::renderUserLink(TextNode* n)
{
    // Map user name to user Id
    string_t userId = User::getUserIdFromLogin(db, n->text);
    if (userId.empty()) {
        return false;
    }
    User u(db, userId);

    // Build it
    result += "<a href=\"";
    renderText(state.getBaseUrl());
    result += USER_BASE_URL;
    renderText(u.getLoginName());
    result += "\">";
    if (n->children.empty()) {
        renderText(u.getScreenName());
    } else {
        renderChildren(n);
    }
    result += "</a>";
    return true;
}

bool
HtmlRenderer::renderPostLink(TextNode* n)
{
    // Validate
    int32_t messageId;
    if (!stringValue(n->text, messageId) || messageId <= 0) {
        return false;
    }

    // Access thread, check permissions
    Message m(db, messageId);
    if (!m.exists()) {
        // FIXME permission check!
        return false;
    }
    int32_t topicId = m.topicId().get();
    Topic t(db, topicId);

    // OK, we are allowed to access it. Get its name.
    const string_t postName = m.subject().get();
    const string_t topicName = t.subject().get();
    result += "<a href=\"";
    renderText(state.getBaseUrl());
    result += makePostUrl(topicId, topicName, messageId);
    result += "\">";
    if (n->children.empty()) {
        if (postName.empty()) {
            renderText("(no subject)");
        } else {
            renderText(abbreviate(postName));
        }
    } else {
        renderChildren(n);
    }
    result += "</a>";
    return true;
}

bool
HtmlRenderer::renderThreadLink(TextNode* n)
{
    // Validate
    int32_t topicId;
    if (!stringValue(n->text, topicId) || topicId <= 0) {
        return false;
    }

    // Access thread, check permissions
    Topic t(db, topicId);
    if (!t.exists()) {
        // FIXME permission check!
        return false;
    }

    // OK, we are allowed to access it. Get its name.
    const string_t name = t.subject().get();
    result += "<a href=\"";
    renderText(state.getBaseUrl());
    result += makeTopicUrl(topicId, name);
    result += "\">";
    if (n->children.empty()) {
        renderText(abbreviate(name));
    } else {
        renderChildren(n);
    }
    result += "</a>";
    return true;
}

bool
HtmlRenderer::renderImage(TextNode* n)
{
    // Validate
    if (!isValidUrl(n->text)) {
        return false;
    }

    // Render
    result += "<img src=\"";
    renderText(n->text);
    if (!n->children.empty()) {
        result += "\" alt=\"";
        renderText(n->rawTextContent());
    }
    result += "\" />";
    return true;
}


void
HtmlRenderer::renderChildren(TextNode* n)
{
    for (ptr_vector<TextNode>::size_type i = 0, e = n->children.size(); i != e; ++i) {
        render(n->children[i]);
    }
}

void
HtmlRenderer::render(TextNode* n)
{
    bool ok;
    const char* pfx;
    int i;
    switch (n->major) {
     case maPlain:
        renderText(n->text);
        break;

     case maInline:
        if (const char* tag = getInlineTagName(n->minor)) {
            result += "<";
            result += tag;
            result += ">";
            renderChildren(n);
            result += "</";
            result += tag;
            result += ">";
        } else {
            renderChildren(n);
        }
        break;

     case maInlineAttr:
        switch (n->minor) {
         case miIAColor:
            result += "<font color=\"";
            renderText(n->text);
            result += "\">";
            renderChildren(n);
            result += "</font>";
            break;

         case miIAFont:
            result += "<span style=\"font-family: ";
            renderText(n->text);
            result += ";\">";
            renderChildren(n);
            result += "</span>";
            break;

         case miIASize:
            if (stringValue(n->text, i) && i >= -8 && i <= +8 && i != 0) {
                int32_t size = 100;
                while (i > 0) {
                    size *= 100;
                    size /= 80;
                    --i;
                }
                while (i < 0) {
                    size *= 80;
                    size /= 100;
                    ++i;
                }
                result += format("<span style=\"font-size: %d%%\">", size);
                renderChildren(n);
                result += "</span>";
            } else {
                renderChildren(n);
            }
            break;

         default:
            renderChildren(n);
        }
        break;

     case maLink:
        switch (n->minor) {
         case miLinkUrl:
            ok = isValidUrl(n->text) && renderLink(n, "<a href=\"", "\" rel=\"nofollow\">", "</a>");
            pfx = "link ";
            break;
         case miLinkEmail:
            ok = renderLink(n, "<a href=\"mailto:", "\">", "</a>");
            pfx = "mail ";
            break;
         case miLinkThread:
            ok = renderThreadLink(n);
            pfx = "thread ";
            break;
         case miLinkPost:
            ok = renderPostLink(n);
            pfx = "post ";
            break;
         case miLinkGame:
            ok = renderGameLink(n);
            pfx = "game ";
            break;
         case miLinkUser:
            ok = renderUserLink(n);
            pfx = "user ";
            break;
         case miLinkForum:
            ok = renderForumLink(n);
            pfx = "forum ";
            break;
         case miLinkImage:
            ok = renderImage(n);
            pfx = "image ";
            break;
         default:
            pfx = "";
            ok = false;
            break;
        }
        if (!ok) {
            result += "<span class=\"tfailedlink\">";
            if (n->children.empty()) {
                renderText(pfx);
                renderText(n->text);
            } else {
                renderChildren(n);
            }
            result += "</span>";
        }
        break;

     case maParagraph:
        switch (n->minor) {
         case miParNormal:
            result += "<p>";
            renderChildren(n);
            result += "</p>\n";
            break;
         case miParCode:
            result += "<pre>";
            renderChildren(n);
            result += "</pre>\n";
            break;
         case miParCentered:
            result += "<center>";
            renderChildren(n);
            result += "</center>\n";
            break;
         case miParBreak:
            break;
         default:
            renderChildren(n);
            break;
        }
        break;
     case maGroup:
        switch (n->minor) {
         case miGroupQuote:
            if (!n->text.empty()) {
                result += "<div class=\"attribution\">";
                renderAttribution(n->text);
                result += "</div>\n";
            }
            result += "<blockquote>";
            renderChildren(n);
            result += "</blockquote>\n";
            break;
         case miGroupList:
            {
                const char* close;
                if (n->text.empty()) {
                    result += "<ul>";
                    close = "</ul>";
                } else if (n->text == "1") {
                    result += "<ol>";
                    close = "</ol>";
                } else {
                    result += "<ol type=\"";
                    renderText(n->text);
                    result += "\">";
                    close = "</ol>";
                }
                if (n->isSimpleList()) {
                    for (ptr_vector<TextNode>::size_type i = 0, e = n->children.size(); i != e; ++i) {
                        result += "<li>";
                        renderChildren(n->children[i]->children[0]);
                        result += "</li>\n";
                    }
                } else {
                    renderChildren(n);
                }
                result += close;
            }
            break;
         case miGroupListItem:
            result += "<li>";
            renderChildren(n);
            result += "</li>\n";
            break;
         case miGroupRoot:
         default:
            renderChildren(n);
            break;
        }
    }
}

string_t
renderHTML(TextNode* n, RenderState& state, DbClient& db)
{
    string_t result;
    HtmlRenderer(state, db, result).render(n);
    return result;
}
